#!/bin/sh
# the next line restarts using wish \
exec wish8 "$0" -visual truecolor "$@"

#  qsmackfe.tcl -- http://lemur.stanford.edu/clan9/qsmack
#
#  Written By: Jeff Solomon (jsolomon@stanford.edu)
#
# *******************************************************************
# *                                                                 *
# *               Copyright (C) 1997, 1998 Jeff Solomon             *
# *                                                                 *
# *   These coded instructions, statements, and computer programs   *
# *   are the property of Jeff Solomon and are protected by         *     
# *   Federal copyright law. This program may be freely modified    *
# *   as long as this notice is kept intact, but may not be         *
# *   redistributed to third parties in any form, in whole or in    *
# *   part, without the prior written consent of Jeff Solomon.      *
# *                                                                 *
# *******************************************************************

# File Organization - the procedures in this file are separated into 3 distincts
# types. 
# 
# Those in all lower case are procedures that are called when the user types in
# a command, a command is 'stuffed' from the QSmack server, or exec'ed from a
# file.
#
# Those procedures with First letter uppercase and mixed case, are internal only
# functions. They are only called from other functions within the frontend.
#
# All uppercase procedures are called as a result of receiving a packet from
# QSmack. When a packet is received, the first word of the received message is
# converted to all upper case and then the packet is executed. See the
# 'RecvPacket' procedure for details.
# 

# What the hell is this doing? - I'll tell ya
# The variables cfg(titleFont), cfg(infoFont), and cfg(textFont) have to be
# defined. As a default I want them to be defined to what value Tk would choose
# on its own. So I create a dummy label and entry and get the font from those.
label .dummylabel
entry .dummyentry
set cfg(titleFont)               [.dummylabel cget -font]
set cfg(infoFont)                [.dummylabel cget -font]
set cfg(textFont)                [.dummyentry cget -font]
destroy .dummylabel
destroy .dummyentry

set cfg(password)                ""
set cfg(defaultQSmackServer)     ""
set cfg(width)                   80
set cfg(historyDepth)            50
set cfg(saveTextLines)           500
set cfg(serverWindowHeight)      4
set cfg(smackWindowHeight)       4
set cfg(commandWindowHeight)     4
set cfg(serverPrints)            server
set cfg(smackPrints)             smack
set cfg(commandPrints)           command
set cfg(showNum)                 true
set cfg(nameGfx)                 true
set cfg(concharsFile)            "conchars.gif"
set cfg(showName)                true
set cfg(showColors)              true
set cfg(showFrags)               true
set cfg(showIpaddr)              false

set colors(0)                    White
set colors(1)                    Brown
set colors(2)                    LightBlue
set colors(3)                    DarkKhaki
set colors(4)                    Red
set colors(5)                    SandyBrown
set colors(6)                    Salmon
set colors(7)                    PeachPuff4
set colors(8)                    Purple4
set colors(9)                    Purple1
set colors(10)                   Tan
set colors(11)                   ForestGreen
set colors(12)                   Yellow
set colors(13)                   Blue

# To add a new command, all you have to do is add the command name to this list
# (so the frontend knows about it and command completion will work), and then
# write the routine that will called when the user types in that command. The
# name of the procedure will be the same as the command, and the arguments
# passed to the command will be one string that contains the rest of the
# command.
set commands {
                                 alias
                                 ban
                                 ban_ip
                                 ban_name
                                 ban_color
                                 ban_exclude
                                 clear
                                 color
                                 connect
                                 disconnect
                                 exec
                                 exit
                                 fire
                                 help
                                 history
                                 impulse
                                 jump
                                 kick
                                 lockban
                                 masquerade
                                 mask
                                 name
                                 nslookup
                                 pause
                                 ping
                                 quake
                                 qsmack_command
                                 qsmack_connect
                                 qsmack_disconnect
                                 qsmack_quit
                                 quit
                                 reread_banfile
                                 say
                                 say_banlist
                                 say_team
                                 say_qsmack
                                 servername
                                 stagename
                                 status
                                 tell
                                 unalias
                                 unban
                                 unlockban
                                 who
}

set ignorecommands {
                                 ""
                                 ;
                                 bind
                                 init
                                 play
                                 reconnect
                                 wait
}

proc InitVars { } {
  global q hist statusdirty smackmode masked_as_num commands allcommands cfg

  set q(qsmackserver)              ""
  set q(qsmackport)                ""
  set q(qserver)                   ""
  set q(qport)                     ""
  set q(qmapname)                  ""
  set q(qmaxplayers)               0
  set q(stagename)                 ""
  set q(raw_stagename)             ""
  set q(playername)                ""
  set q(playernumber)              0
  set q(playershirt)               0
  set q(playerpants)               0
  set q(qtime)                     "00:00:00"
  set q(version)                   ""

  set hist(index)                  0
  set hist(end)                    0
  set hist(depth)                  $cfg(historyDepth)
  history keep                     $cfg(historyDepth)

  set statusdirty                  0
  set smackmode                    0
  set masked_as_num                0
  set allcommands                  [lsort $commands]

  wm iconname . "qsmack"
}

proc LoadQSmackExtras { } {
  global command qsmack_dynlib cfg argv0

  set so [info sharedlibextension]
  regexp {^(.*)qsmackfe.tcl$} $argv0 dummy path
  if { $path == "" } {
    $path = "./"
  }
  if { $so != "" } {
    set qsmack_dynlib "${path}libtcl_qsmackfe$so"
    if { [catch {load "libtcl_qsmackfe$so"} reason] } {
      if { [catch {load "${path}libtcl_qsmackfe$so"} reason] } {
        Scroll $command "Couldn't load 'libtcl_qsmackfe$so'\n"
        Scroll $command "Couldn't load '${path}libtcl_qsmackfe$so'\n"
        Scroll $command "No 'nslookup' command for you!\n"
      }
    }
  }

  set concharsFile ${path}$cfg(concharsFile)
  if { $cfg(nameGfx) == "true" } {
    if { [file readable $concharsFile] } {
      image create photo alphabet -file $concharsFile
      for { set y 0 } { $y < 16 } { incr y } {
          for { set x 0 } { $x < 16 } { incr x } {
              set intcode [expr $y*16+$x]
              image create photo QuakeChars($intcode) -width 16 -height 16
              QuakeChars($intcode) copy alphabet \
                  -zoom 2 2 \
                  -from [expr $x*8] [expr $y*8] [expr $x*8+8] [expr $y*8+8] \
                  -to 0 0 16 16
          }
      }
      image delete alphabet
    } else {
      Scroll $command "Couldn't load QuakeChars file: '$concharsFile'\n"
      Scroll $command "No quake font for you!\n"
      set cfg(nameGfx) false
    }
  }
}

proc ResetVars { } {
  global players 

  for {set i 0} {$i < 16} {incr i} {
    set players($i,valid)  0
    set players($i,num)    0
    set players($i,ipaddr) ""
    set players($i,dns)    ""
    set players($i,name)   ""
    set players($i,name_raw)   ""
    set players($i,frags)  0
    set players($i,shirt)  0
    set players($i,pants)  0
  }

  wm title . "QSmack not connected"
}

proc ReadConfigFile { } { 
  global argc argv env cfg initcmds

  set cfgfile ""

  if { $argc == 1 } {
    set cfgfile $argv
  } elseif { [info exists env(PREF_FOLDER)] } {
    set cfgfile "$env(PREF_FOLDER)QSmack Preferences"
  } elseif { [info exists env(HOME)] } {
    set cfgfile "$env(HOME)/.qsmack"
  }

  if { $cfgfile != "" && [file readable $cfgfile] } {
    if { [catch {set fp [open $cfgfile r]} reason] } {
      puts stderr $reason
      exit
    }
    set i 0
    while { [gets $fp line] >= 0 } {
      if { [regexp {^ *#} $line] || [regexp {^[ 	]*$} $line] } {
       continue
      }
      if { [scan $line "%\[^:\]:%*\[ \t\]%\[^\n\]" option value] == 2 } {
        if { $option == "command" } {
          set initcmds($i) $value
          incr i
        } elseif { [info exists cfg($option)] } {
          set cfg($option) $value
        } else {
          puts stderr "Unknown option in '$cfgfile': $option"
        }
      } else {
       puts stderr "Bad line in '$cfgfile': $line"
      }
    }
    close $fp
  }
}

proc Wrap { text } {
  global q

  set r ""
  set len [expr 35 - [string length $q(stagename)]]
 
  while { [string length $text] != 0 } {
    if { [string length $text] <= $len } {
       lappend r $text
       break
    }
    for {set l $len} {$l >= 0} {incr l -1} {
      if { [string index $text $l] == " " } {
        break
      }
    }
    if { $l <= 0 } {
      set l $len
    }
    lappend r [string range $text 0 $l]
    set text [string range $text [expr $l+1] end]
  } 
  return $r
} 

proc GetTokens { cmd } {
  set words ""
  regsub {^[ 	]+} $cmd "" cmd

  while { [string length $cmd] != 0 } {
      set token ""
      if { [regexp {^"(([^"\\]|\\.)*)"} $cmd dummy token] } {
        regsub {^"([^"\\]|\\.)*"} $cmd "" cmd
      } elseif { [regexp {^"} $cmd] } {
          set words "parseError"
          break
      } elseif { [regexp {^(([^ 	;"\\]|\\.)+)} $cmd dummy token] } {
        regsub {^(([^ 	;"\\]|\\.)+)} $cmd "" cmd
      } elseif { [regsub {^;} $cmd "" cmd] } {
          set token ";"
      } else {
          regsub {^[ 	]+} $cmd "" cmd
          continue
      }
      lappend words $token
  }
  return $words
}

proc AddHistory { text } {
  global hist command

  if { [info exists hist([expr $hist(end) - $hist(depth)])] } {
    unset hist([expr $hist(end) - $hist(depth)])
  }
  set  hist($hist(end))   $text
  set  hist(index)        [incr hist(end)]
  set  hist($hist(end))   ""
}

proc UpHistory { } {
  global hist entry command
  
  set previndex [expr $hist(index) - 1]

  if { [info exists hist($previndex)] } {
    set hist(index) $previndex
    $entry.entry delete 0 end
    $entry.entry insert 0 $hist($hist(index))
  }
}

proc DownHistory { } {
  global hist entry command
  
  $entry.entry delete 0 end

  set nextindex [expr ($hist(index) + 1)]

  if { [info exists hist($nextindex)] } {
    set hist(index) $nextindex
    $entry.entry delete 0 end
    $entry.entry insert 0 $hist($hist(index))
  }
}

proc ListCompletions { } {
  global allcommands command entry cfg

  set cmd        [$entry.entry get]
  set insert     [$entry.entry index insert]
  set completion ""
  set matches    ""
  set longest    0
  
  set suffix     [string range $cmd 0 [expr $insert -1]]
  set nextch     [string index $cmd $insert]

  if { [regexp {^ ?$} $nextch] && 
       [regexp {^ *([-a-zA-Z_0-9]*)$} $suffix dummy cmd] } {
    foreach c $allcommands {
      if { [regexp "^$cmd" $c] } {
        lappend matches $c
        if { [string length $c] > $longest } {
          set longest [string length $c]
        }
      }
    }
    set i          [expr $cfg(width) / ($longest + 1)]
    set i          [expr ($i == 0) ? 1 : $i]
    set nummatches [llength $matches]
    set skip       [expr int(ceil(double($nummatches)/$i))]
    for {set j 0} {$j < $skip} {incr j} {
      for {set k 0} {$k < $i} {incr k} {
        set index [expr $j + ($k * $skip)]
        if { $index < $nummatches } {
          Scroll $command [format "%-${longest}s " [lindex $matches $index]]
        }
      }
      Scroll $command "\n"
    }
  }
}

proc CommandCompletion { } {
  global entry allcommands

  set cmd        [$entry.entry get]
  set insert     [$entry.entry index insert]
  set completion ""
  set matches    ""
  
  set suffix     [string range $cmd 0 [expr $insert -1]]
  set nextch     [string index $cmd $insert]

  if { [regexp {^ ?$} $nextch] && 
       [regexp {^ *([-a-zA-Z_0-9]*)$} $suffix dummy cmd] } {
    foreach c $allcommands {
      if { [regexp "^$cmd" $c] } {
        lappend matches $c
      }
    }
    set nummatches [llength $matches]
    if { $nummatches > 1 } {
      set longest 0
      for {set i 1} { $i < [llength $matches] } {incr i} {
        if { [string length [lindex $matches $i]] > 
             [string length [lindex $matches $longest]] } {
          set longest $i
        }
      }
      set nextch ""
      while { $nummatches == [llength $matches] } {
        set completion "$completion$nextch"
        set nextch     [string index [lindex $matches $longest] [string length $cmd]]
        set cmd        "$cmd$nextch"
        set matches    ""
        foreach c $allcommands {
          if { [regexp "^$cmd" $c] } {
            lappend matches $c
          }
        }
      }
    } elseif { $nummatches == 1 } {
      set completion "[string range [lindex $matches 0] [string length $cmd] end] "
    }
  }

  $entry.entry insert $insert $completion
}

proc BangSubstitute { words } {
  global command hist
  set    cmdword 1
  set    bangwords {}

  foreach word $words {
    set bangword $word

    if { $cmdword && $word == "!!" } {
      set bangwords [concat $bangwords [GetTokens $hist([expr $hist(index) - 1])]]
    } elseif { $cmdword && [string index $word 0] == "!" } {
      if { [ catch { set bangword [history event [string range $word 1 end]] } reason ] } {
        Scroll $command "$reason\n"
      } else {
        set bangwords [concat $bangwords [GetTokens $bangword]]
      }
    } else {
      lappend bangwords $bangword
    }

    if { $word == ";" } {
      set cmdword 1
    } else {
      set cmdword 0
    }
  }
  return $bangwords
}

proc UserInput { text } {
  global smackmode entry hist masked_as_num

  if { $smackmode == 1 } {
    say $text
    AddHistory $text
    history add "say $text"
  } elseif { $smackmode == 2 } {
    say_team $text
    AddHistory $text
    history add "say_team $text"
  } elseif { $smackmode == 3 } {
    say_qsmack $text
    AddHistory $text
    history add "say_qsmack $text"
  } elseif { $smackmode == 4 } {
    masquerade $text
    AddHistory $text
    history add "masquerade $text"
  } else {
    UserCommand $text
  }

  $entry.entry delete 0 end
}

proc UserCommand { text } {
  set words [BangSubstitute [GetTokens $text]]

  if { $words != {} } {
    foreach word $words {
      append cmdhist "$word "
    }
    regexp {^(.*) $} $cmdhist dummy cmdhist
    AddHistory $cmdhist
    history add $cmdhist
    ProcessCommand $words
  }
}

proc ExecStuffAliasCommand { text } {
  set words [GetTokens $text]

  ProcessCommand $words
  update idletasks
}

proc ProcessCommand { words } {
  global command aliases stuffed_aliases ignorecommands commands

  while { [llength $words] } {
    set nextcommand ""
    while { [llength $words] && [lindex $words 0] != ";" } {
      lappend nextcommand [lindex $words 0]
      set words [lreplace $words 0 0]
    }
    if { [lindex $words 0] == ";" } {
      set words [lreplace $words 0 0]
    }
    set cmd  [lindex $nextcommand 0]
    set args [join [lrange $nextcommand 1 end]]
    if { $cmd == "parseError" } {
      Scroll $command "Unmatched quote\n"
    } elseif { $cmd == "quit" || $cmd == "exit" } {
      disconnect
      exit
    } elseif { $cmd == "history" } {
      Scroll $command "[history info]\n"
    } elseif { $cmd == "tell" } {
      if { [ catch { qsmack_tell $args } reason ] } {
        Scroll $command "Error in tell: $reason\n"
      }
    } elseif { $cmd == "exec" } {
      if { [ catch { qsmack_exec $args } reason ] } {
        Scroll $command "Error in exec: $reason\n"
      }
    } elseif { [lsearch -exact $commands $cmd] != -1 } {
      if { [ catch { $cmd $args } reason ] } {
        Scroll $command "Error in $cmd: $reason\n"
      }
    } elseif { [info exists aliases($cmd)] } {
      ExecStuffAliasCommand $aliases($cmd)
    } elseif { [info exists stuffed_aliases($cmd)] } {
      ExecStuffAliasCommand $stuffed_aliases($cmd)
    } elseif { [lsearch -exact $ignorecommands $cmd] != -1 } {
      continue
    } else {
      Scroll $command "Unknown command: $cmd\n"
    }
  }
}

proc SendPacket { packet } {
  global socketid command

  if { [info exists socketid] } {
    puts  $socketid "$packet"
    flush $socketid
  } else {
    Scroll $command "Command '$packet' ignored, not connected to QSmack!\n"
    Scroll $command "Use the 'connect' command to connect to a QSmack.\n"
  }
}

proc RecvPacket { } { 
  global socketid command

  set pkt [gets $socketid]

  if { [string length $pkt] == 0 } {
    Scroll $command "zero length command from QSmack, disconnecting\n"
    disconnect
    return
  }

  if { [regexp {^([a-zA-Z_0-9]+) ?(.*)$} $pkt dummy cmd args] == 0 } {
    Scroll $command "bad command from qsmack server: $pkt\n"
    return
  }

  if { [ catch { [string toupper $cmd] $args } reason ] } {
    Scroll $command "$reason: $pkt\n"
  }

}

proc connect { arg } {
  global socketid q command

  if { [regexp {^([-a-zA-Z0-9_.]+):([0-9]+)$} $arg dummy server port] } {
    disconnect
    set socketid [socket $server $port]
    fconfigure $socketid -translation {auto lf}
    fileevent  $socketid readable { RecvPacket }
    set q(qsmackserver) $server
    set q(qsmackport)   $port
  } else {
    Scroll $command "bad arg to 'connect': $arg\n"
  }
}

proc disconnect { args } {
  global socketid stuffed_aliases allcommands command

  if { [info exists socketid] } {
    SendPacket "disconnect"
    close $socketid
    unset socketid
    foreach alias [array names stuffed_aliases] {
      set index       [lsearch -exact $allcommands $alias]
      set allcommands [lreplace $allcommands $index $index]
    }
    unset stuffed_aliases
    ResetVars
  }
}

proc alias { arg } {
  global command aliases stuffed_aliases allcommands

  if { [llength $arg] == 0 } {
    if { [array size aliases] > 0 } {
      Scroll $command "Currently defined user aliases:\n"
      foreach alias [lsort [array names aliases]] {
          Scroll $command [format "user:    %-20s%s\n" $alias $aliases($alias)]
      }
      Scroll $command "\n"
    } else {
      Scroll $command "No user aliases currently defined\n"
    }
    if { [array size stuffed_aliases] > 0 } {
      Scroll $command "Currently defined stuffed aliases:\n"
      foreach alias [lsort [array names stuffed_aliases]] {
          Scroll $command [format "stuffed: %-20s%s\n" $alias $stuffed_aliases($alias)]
      }
    } else {
      Scroll $command "No stuffed aliases currently defined\n"
    }
  } elseif { [llength $arg] == 1 } {
    set alias [lindex $arg 0]
    if { [info exists aliases($alias)] } {
      Scroll $command [format "user:    %-20s%s\n" $alias $aliases($alias)]
    } elseif { [info exists stuffed_aliases($alias)] } {
      Scroll $command [format "stuffed: %-20s%s\n" $alias $stuffed_aliases($alias)]
    } else {
      Scroll $command "'$alias' not aliased\n"
    }
  } else {
    set alias           [lindex $arg 0]
    set what            [join [lrange $arg 1 end]]
    if { [lindex [info level 2] 0] == "STUFF" } {
      if { [info exists stuffed_aliases($alias)] == 0 } {
        lappend allcommands $alias
        set allcommands      [lsort $allcommands]
      }
      set stuffed_aliases($alias) $what
    } else {
      if { [info exists aliases($alias)] == 0 } {
        lappend allcommands $alias
        set allcommands      [lsort $allcommands]
      }
      set aliases($alias) $what
    }

  }
}

proc ban { arg } {
  global players command

  if { $arg == "" } {
    SendPacket "banlist"
  } elseif { [scan $arg "%d" num] == 1 } {
     incr num -1
     if { $num >= 0 && $num < 16 && $players($num,valid) } {
       SendPacket "ban_ip $players($num,ipaddr) $players($num,name)\n"
     } else {
       Scroll $command "Invalid player number: [incr num]\n"
     }
  } else {
    Scroll $command "Bad arg to 'ban': $arg\n"
  }
}

proc ban_color { arg } {
  global command

  if { [scan $arg "%d" shirt] } {
    if { [scan $arg "%d %d" shirt pants] != 2 } {
      set pants $shirt
    }
    if { $shirt >= 0 && $shirt <= 13 && $pants >=0 && $pants <= 13 } {
      SendPacket "ban_color $shirt $pants"
    } else {
      Scroll $command "Bad ban_color values: $shirt $pants\n"
    }
  } else {
    Scroll $command "Bad args to 'ban_color': $arg\n"
  }
}

proc ban_name { arg } {
  global command

  if { $arg == "" } {
    SendPacket "banlist"
  } else { 
    SendPacket "ban_name $arg"
  }
}

proc ban_ip { arg } {
  global command

  if { $arg == "" } {
    SendPacket "banlist"
  } else { 
    SendPacket "ban_ip $arg"
  }
}

proc ban_exclude { arg } {
  global command

  if { $arg == "" } {
    SendPacket "banlist"
  } else { 
    SendPacket "ban_exclude $arg"
  }
}

proc say_banlist { arg } {
  SendPacket "say_banlist"
}

proc color { arg } {
  global command

  if { [scan $arg "%d" shirt] } {
    if { [scan $arg "%d %d" shirt pants] != 2 } {
      set pants $shirt
    }
    if { $shirt >= 0 && $shirt <= 13 && $pants >=0 && $pants <= 13 } {
      SendPacket "color $shirt $pants"
    } else {
      Scroll $command "Bad color values: $shirt $pants\n"
    }
  } else {
    Scroll $command "Bad args to 'color': $arg\n"
  }
}

proc clear { arg } {
  global server smack command

  if { $arg == "" } {
    Clear $server
    Clear $smack
    Clear $command
  } elseif { $arg == "server" } {
    Clear $server
  } elseif { $arg == "smack" } {
    Clear $smack
  } elseif { $arg == "command" } {
    Clear $command
  } else {
    Scroll $command "Bad args to 'clear': $arg\n"
  }
}

proc qsmack_exec { execfile } { 
  global command

  if { $execfile != "" && [file readable $execfile] } {
    set fp [open $execfile r]
    while { [gets $fp line] >= 0 } {
      if { [regexp {^ *#} $line] || [regexp {^[ 	]*$} $line] } {
       continue
      }
      ExecStuffAliasCommand $line
    }
    close $fp
  } else {
    Scroll $command "Couldn't open file: $execfile\n"
  }
}

proc help { arg } {
  global command
  Scroll $command "What kind of stupid retard needs help? Screw off!\n"
}

proc fire { arg } {
  SendPacket "fire"
}

proc jump { arg } {
  SendPacket "jump"
}

proc impulse { arg  } {
  global command

  if { [scan $arg "%d" impulse] == 1 &&
       $impulse >=0 && $impulse < 256 } {
    SendPacket "impulse $impulse"
  } else {
    Scroll $command "Bad arg to 'impulse': $arg\n"
  }
}

proc kick { arg } {
  global players command

  foreach n [lsort $arg] {
    if { [scan $n "%d" num] == 1 } {
       incr num -1
       if { $num >=  0 && $num <  16 && $players($num,valid) } {
         SendPacket "kick [incr num]"
       } else {
         Scroll $command "Invalid player number: [incr num]\n"
       }
    } else {
      Scroll $command "Bad arg to 'kick': $n\n"
    }
  }
}

proc lockban { arg } {
  SendPacket "lockban"
}

proc mask { arg } {
  masquerade $arg
}

proc masquerade { arg } {
  global command masked_as_num statusdirty players

  if { [scan $arg "%d" num] == 1 } {
    incr num -1
    if { $players($num,valid) } {
      set masked_as_num   [expr $num + 1]
      set what            [join [lrange $arg 1 end]]
      set statusdirty     1
    } else {
      Scroll $command "Invalid player number: [incr num]\n"
      return
    }
  } else {
    set what            [join [lrange $arg 0 end]]
  }

  set lines [Wrap $what]
  foreach line $lines {
    SendPacket "masquerade $masked_as_num $line"
  }
}

proc pause { arg } {
  SendPacket "quake pause"
}

proc ping { arg } {
  SendPacket "quake ping"
}

proc quake { arg } {
  SendPacket "quake $arg"
}

proc qsmack_command { arg } {
  SendPacket "$arg"
}

proc qsmack_quit { arg  } {
  SendPacket "qsmack_quit"
}

proc qsmack_connect { arg  } {
  SendPacket "qsmack_connect $arg"
}

proc qsmack_disconnect { arg  } {
  SendPacket "qsmack_disconnect"
}

proc reread_banfile { arg } {
  SendPacket "reread_banfile"
}

proc servername { name } {
  SendPacket "servername $name"
}

proc name { name } {
  stagename $name
}

proc stagename { name } {
  global q socketid

  if { [info exists socketid] } {
    set q(raw_stagename) $name
  }
  SendPacket "stagename $name"
}

# Doing a gethostbyaddr can take a few seconds, sometimes up to 10. When doing a
# lookup on all players, update the screen after every lookup so the user knows
# the frontend is still alive.
proc nslookup { arg } {
  global command players qsmack_dynlib

  if { [info commands gethostbyaddr] == "gethostbyaddr" } {
    if { $arg == "" } {
      for {set i 0} {$i < 16} {incr i} {
        if { $players($i,valid) } {
          _nslookup $i
          update idletasks
        }
      }
    } elseif { [scan $arg "%d" num] == 1 } {
       incr num -1
       if { $num >= 0 && $num < 16 && $players($num,valid) } {
         _nslookup $num
       } else {
         Scroll $command "Invalid player number: [incr num]\n"
       }
    } else {
      Scroll $command "Bad arg to 'nslookup': $arg\n"
    }
  } else {
    Scroll $command "'$qsmack_dynlib' module could not be loaded\n"
  }
}

# If we have already looked up this player's DNS entry before, use the saved
# version. No need to wait twice.
proc _nslookup { num } {
  global command players

  if { $players($num,dns) == "" } {
    set players($num,dns) [gethostbyaddr $players($num,ipaddr)]
  }
  Scroll $command [format "#%-2d %-15s : $players($num,dns)\n" \
                    [expr $num+1] $players($num,name)]
}

proc status { arg } {
  SendPacket "status"
}

proc say_qsmack { what } {
  set lines [Wrap $what]
  foreach line $lines {
    SendPacket "say_qsmack $line"
  }
}

proc say_team { what } {
  set lines [Wrap $what]
  foreach line $lines {
    SendPacket "say_team $line"
  }
}

proc say { what } {
  set lines [Wrap $what]
  foreach line $lines {
    SendPacket "say $line"
  }
}

proc qsmack_tell { what } {
  global command players

  if { [scan $what "%d%*c%\[^\n\]" who text] == 2 } {
    if { $who > 0 && $who <= 16 && $players([expr $who-1],valid) } {
      set lines [Wrap $text]
      foreach line $lines {
        SendPacket "tell $who $line"
      }
    } else {
      Scroll $command "Invalid player number for 'tell': $who\n"
    }
  } else {
    Scroll $command "Bad args to 'tell': $what\n"
  }
}

proc unalias { alias } {
  global command aliases stuffed_aliases allcommands

  if { $alias != "" } {
     if { [info exists aliases($alias)] } {
       unset aliases($alias)
       if { [set index [lsearch -exact $allcommands $alias]] != -1 } {
         set allcommands [lreplace $allcommands $index $index]
       }
     } elseif { [info exists stuffed_aliases($alias)] } {
       unset stuffed_aliases($alias)
       if { [set index [lsearch -exact $allcommands $alias]] != -1 } {
         set allcommands [lreplace $allcommands $index $index]
       }
     } else {
       Scroll $command "Not aliased: $alias\n"
     }
  } else {
     foreach alias [array names aliases] {
       set index       [lsearch -exact $allcommands $alias]
       set allcommands [lreplace $allcommands $index $index]
     }
     unset aliases
     foreach alias [array names stuffed_aliases] {
       set index       [lsearch -exact $allcommands $alias]
       set allcommands [lreplace $allcommands $index $index]
     }
     unset stuffed_aliases
     Scroll $command "Removed all aliases\n"
  } 
} 

proc unban { arg } {
  global command
  
  foreach n [lsort -integer -decreasing $arg] {
    if { [scan $n "%d" num] == 1 } {
      SendPacket "unban $n"
    } else {
      Scroll $command "Bad arg to 'unban': $n\n"
    }
  }
}    

proc unlockban { arg } {
  SendPacket "unlockban"
}

proc who { arg } {
  SendPacket "who"
}

proc PASSWORD { arg } {
  global cfg command
  
  if { $cfg(password) == "" } {
    Scroll $command "You must define a password to send to the qsmack server\n"
    Scroll $command "Add a 'password' option to your qsmack.tcl startup file\n"
    disconnect
  } else {
    SendPacket "password $cfg(password)"
  }
}

proc PRINT { text } {
  global server

  Scroll $server "$text\n"
  update idletasks
}
  
proc SAYPRINT { text } {
  global smack

  Scroll $smack "$text\n"
  update idletasks
}

proc COMMANDPRINT { text } {
  global command

  Scroll $command "$text\n"
  update idletasks
}

proc ACCEPTED { arg } {
  global env q

  if { [info exists env(USER)] } {
    set username $env(USER)
  } else {
    set username user
  }

  scan [info hostname] {%[^.]} hostname
 
  if { $q(raw_stagename) != "" } {
    SendPacket "stagename $q(raw_stagename)"
  }
  SendPacket "realname $username@$hostname"
}

proc EXCEPTION { text } {
  global command

  Scroll $command "$text\n"
}

proc MAPNAME { map } {
  global q command

  set q(qmapname) $map
  wm title . "QSmack: $q(qsmackserver):$q(qsmackport) Quake: $q(qserver) - $q(qmapname)"
}

proc MAXPLAYERS { num } {
  global q

  set q(qmaxplayers) $num
}

proc MYSTAGENAME { name } {
  global q

  set q(stagename) $name
}

proc MYNAME { name } {
  global q

  set q(playername) $name
}

proc MYNUMBER { num } {
  global q

  set q(playernumber) $num
}

proc MYSHIRT { shirtcolor } {
  global q colors statusbar

  set q(playershirt) $colors($shirtcolor)
  $statusbar.colors.shirt configure -bg $q(playershirt)
}

proc MYPANTS { pantscolor } {
  global q colors statusbar

  set q(playerpants) $colors($pantscolor)
  $statusbar.colors.pants configure -bg $q(playerpants)
}

proc PORT { port } {
  global q

  set q(qport) $port
}

proc REJECTED { why } {
  global command

  Scroll $command "$why\n"
  disconnect
}

proc SERVER { servername } {
  global q

  set q(qserver) $servername
}

proc STUFF { stuffcommand } {
  ExecStuffAliasCommand $stuffcommand
}

proc TIME { qtime } {
  global q

  set q(qtime) [format "%02d:%02d:%02d" [expr int($qtime)/3600] \
              [expr (int($qtime)%3600)/60]  [expr int($qtime)%60] ]
  UpdateStatus
}

proc VERSION { version } {
  global q

  set q(qversion) $version
}

proc PLAYEREXIT { arg } {
  global players statusdirty command q

  if { [scan $arg "%d" num ] == 1 } {
    set players($num,valid)  0
    set players($num,dns)    ""
    set statusdirty          1
  } else {
    Scroll $command "Bad 'playerexit' packet from QSmack: $arg\n"
  }

  set num 0
  for {set i 0} {$i < $q(qmaxplayers)} {incr i} {
    if { $players($i,valid) == 1 } {
      incr num
    }
  }
  wm title . "QSmack: $q(qsmackserver):$q(qsmackport) Quake: $q(qserver) - $q(qmapname) \($num/$q(qmaxplayers)\)"
}

proc PLAYERIP { arg } {
  global players statusdirty command q

  if { [scan $arg "%d %s" num ipaddr] == 2 } {
    set players($num,valid)  1
    set players($num,num)    [expr $num+1]
    set players($num,ipaddr) $ipaddr
    set statusdirty          1
  } else {
    Scroll $command "Bad 'playerip' packet from QSmack: $arg\n"
  }

  set num 0
  for {set i 0} {$i < $q(qmaxplayers)} {incr i} {
    if { $players($i,valid) == 1 } {
      incr num
    }
  }
  wm title . "QSmack: $q(qsmackserver):$q(qsmackport) Quake: $q(qserver) - $q(qmapname) \($num/$q(qmaxplayers)\)"
}

proc PLAYERNAME { arg } {
  global players statusdirty

  regexp {^[ 	]*([0-9]+)[ 	](.*)$} $arg dummy num name

  set players($num,name)   $name
  set statusdirty          1
}

proc PLAYERNAME_RAW { arg } {
  global players statusdirty

  regexp {^[ 	]*([0-9]+)[ 	](.*)$} $arg dummy num name

  set players($num,name_raw)   $name
  set statusdirty          1
}

proc PLAYERFRAGS { arg } {
  global players statusdirty command

  if { [scan $arg "%d %d" num frags] == 2 } {
    set players($num,frags)  $frags
    set statusdirty          1
  } else {
    Scroll $command "Bad 'playerfrags' packet from QSmack: $arg\n"
  }
}

proc PLAYERSHIRT { arg } {
  global players statusdirty command

  if { [scan $arg "%d %d" num shirt] == 2 } {
    set players($num,shirt)  $shirt
    set statusdirty          1
  } else {
    Scroll $command "Bad 'playershirt' packet from QSmack: $arg\n"
  }
}

proc PLAYERPANTS { arg } {
  global players statusdirty command

  if { [scan $arg "%d %d" num pants] == 2 } {
    set players($num,pants)  $pants
    set statusdirty          1
  } else {
    Scroll $command "Bad 'playerpants' packet from QSmack: $arg\n"
  }
}

proc StatusSort { a b } {
  global players 

  if { $players($b,valid) == 0 } {
    return -1
  } elseif { $players($a,valid) == 0 } {
    return 1
  } elseif { $players($a,frags) > $players($b,frags) } {
    return -1
  } elseif { $players($a,frags) < $players($b,frags) } {
    return 1
  } elseif { $players($a,num) > $players($b,num) } {
    return 1
  } else {
    return -1
  }
}

proc DrawPlayerNameGfx { c name } {
  set tag "TAG$c"
  $c delete $tag

  set i 0
  foreach intcode $name {
    $c create image [expr $i * 16] 0 -tag $tag \
        -anchor nw -image QuakeChars($intcode)
    incr i
  }
}

proc UpdateStatus { } {
  global statusdirty players showplayers colors defbg q cfg masked_as_num
 
  if { $statusdirty } {
    set sortedplayers [lsort -command StatusSort {0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15}]

    for {set i 0} {$i < 16} {incr i} {
      set j [lindex $sortedplayers $i]
      if { $players($j,valid) } {

        if { [expr $j + 1] == $masked_as_num } {
          set showplayers($i,mask) "m"
        } else {
          set showplayers($i,mask) ""
        }
        if { [expr $j + 1] == $q(playernumber) } {
          set showplayers($i,me) "*"
        } else {
          set showplayers($i,me) ""
        }
        set showplayers($i,num)  "#$players($j,num)"
        if { $cfg(nameGfx) == "true" } {
          DrawPlayerNameGfx $showplayers($i,name) $players($j,name_raw)
        } else {
          set showplayers($i,name)   $players($j,name)
        }
        set showplayers($i,frags)  $players($j,frags)
        set showplayers($i,ipaddr) $players($j,ipaddr)
        $showplayers($i,shirt) configure -bg $colors($players($j,shirt))
        $showplayers($i,pants) configure -bg $colors($players($j,pants))
      } else {
        set showplayers($i,mask)   ""
        set showplayers($i,me)     ""
        set showplayers($i,num)    ""
        if { $cfg(nameGfx) == "true" } {
          DrawPlayerNameGfx $showplayers($i,name) ""
        } else {
          set showplayers($i,name) ""
        }
        set showplayers($i,frags)  ""
        set showplayers($i,ipaddr) ""
        $showplayers($i,shirt) configure -bg $defbg
        $showplayers($i,pants) configure -bg $defbg
      }
    }
    set statusdirty 0
  }
}

proc ShirtPantsWidget { parent } {
  frame  $parent.top   -height 2p -width 1c
  pack   $parent.top   -side top -fill y

  frame  $parent.shirt
  pack   $parent.shirt -fill both -expand 1

  frame  $parent.pants
  pack   $parent.pants -fill both -expand 1

  frame  $parent.bot   -height 2p -width 1c
  pack   $parent.bot   -side bottom -fill y
}
 
proc StatusBarWidget { parent } {
  global q cfg

  label  $parent.stagelbl  -text "Stagename:" -anchor w \
                           -font $cfg(infoFont)
  pack   $parent.stagelbl  -pady 0 -ipady 0 -padx 0 -ipadx 0 \
                           -side left -fill x -expand 0
  label  $parent.stagename -textvariable q(stagename) -anchor w -width 21 \
                           -font $cfg(infoFont)
  pack   $parent.stagename -pady 0 -ipady 0 -padx 0 -ipadx 0 \
                           -side left -fill x -expand 1

  frame  $parent.colors
  ShirtPantsWidget $parent.colors
  pack   $parent.colors    -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left -fill both

  label  $parent.namelbl   -text "Servername:" -anchor w -font $cfg(infoFont)
  pack   $parent.namelbl   -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left

  label  $parent.name      -textvariable q(playername) -anchor w -width 21 \
                           -font $cfg(infoFont)
  pack   $parent.name      -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left -fill x -expand 1

  label  $parent.qtime     -textvariable q(qtime) -width 8 -anchor e \
                           -font $cfg(infoFont)
  pack   $parent.qtime     -pady 0 -ipady 0 -padx 0 -ipadx 0 \
                           -side right -fill x -expand 0
}

proc PlayerInfoWidget { parent num } {
  global cfg showplayers

  frame $parent.dummyleft
  pack  $parent.dummyleft -side left -expand 1 -fill x

  label $parent.mask  -width 1 -anchor w -font $cfg(infoFont) \
                      -textvariable showplayers($num,mask)
  label $parent.me    -width 1 -anchor w -font $cfg(infoFont) \
                      -textvariable showplayers($num,me)
  label $parent.num   -width 4 -anchor w -font $cfg(infoFont) \
                      -textvariable showplayers($num,num)
  if { $cfg(showNum) == "true" } {
    pack $parent.mask -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
    pack $parent.me   -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
    pack $parent.num  -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
  }

  if { $cfg(nameGfx) == "true" } {
    canvas $parent.name -width 240 -height 16
    set showplayers($num,name) $parent.name
  } else {
    label $parent.name  -width 21 -anchor w -font $cfg(infoFont) \
          -textvariable showplayers($num,name)
  }
 
  if { $cfg(showName) == "true" } {
    pack $parent.name -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
  }

  label $parent.dummyspace0 -width 1 -font $cfg(infoFont)
  frame $parent.colors
  ShirtPantsWidget $parent.colors
  set showplayers($num,shirt) $parent.colors.shirt
  set showplayers($num,pants) $parent.colors.pants
  if { $cfg(showColors) == "true" } {
    pack $parent.dummyspace0 -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
    pack $parent.colors -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left -fill both
  }
   
  label $parent.frags -width 4 -anchor e -font $cfg(infoFont) \
                      -textvariable showplayers($num,frags)
  if { $cfg(showFrags) == "true" } {
    pack $parent.frags -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
  }
   
  label $parent.dummyspace1 -width 1 -font $cfg(infoFont)
  label $parent.ipaddr -width 15 -anchor w -font $cfg(infoFont) \
                       -textvariable showplayers($num,ipaddr)
  if { $cfg(showIpaddr) == "true" } {
    pack $parent.dummyspace1 -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
    pack $parent.ipaddr -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left
  }
   
  frame $parent.dummyright
  pack  $parent.dummyright -side right -expand 1 -fill x
}

proc PlayerInfoWidgets { parent } {

  for {set i 0} {$i < 2} {incr i} {
    frame $parent.i$i
    pack  $parent.i$i -side left -fill x -expand 1
  }
  for {set i 0} {$i < 2} {incr i} {
    for {set j 0} {$j < 8} {incr j} {
      set  k [expr ($i*8)+$j]
      frame $parent.i$i.j$j
      PlayerInfoWidget  $parent.i$i.j$j $k
      pack $parent.i$i.j$j -side top -fill x
    }
  }
}

proc Clear { w } {
  $w.text configure -state normal
  $w.text delete 1.0 end
  $w.text configure -state disabled
}

proc Scroll { w text } {
  global cfg

  $w.text configure -state normal
  set bottom [lindex [$w.text yview] 1]
  set lines  [$w.text index end]
  $w.text insert end $text

  if { $lines > $cfg(saveTextLines) } {
    $w.text delete 1.0 2.0
  }

  if { $bottom == 1.0 } {
    $w.text see end
  }

  $w.text configure -state disabled
}

# This is the most hideous piece of code ever written. The goal of this proc is
# to find the width of the text output areas in units of characters. Simple
# right? WRONG! Tk tells me the width in pixels, but I don't know the width of
# each character, so...
#       1) turn wrapping off, make sure the bottom of the window is visible
#       2) start adding characters on a blank line (the last line actually)
#       3) do this until they are all not visible
#       4) the number of characters you added while they were all visible is the
#          width (it doesn't matter which char you add since the font should be
#          fixed width.
#       5) Oh yeah, don't forget to save the y-position of this window, because
#          you have to make sure the line you're adding chars to is visible or
#          else the xview will never change and you'll go into an infinite loop.
#       6) Finally, delete the line you just added, adjust the view to what it
#          used to be, and turn word wrapping back on, WOW is this ugly!
proc FigureOutWidth { w } {
  global cfg

  set i 0
  set yview_top [lindex [$w yview] 0]
  set line [expr [$w index end] - 1]
  $w configure -wrap none -state normal
  $w see end
  while { [lindex [$w xview] 1] == 1.0 } {
    $w insert end "O"
    incr i
  }
  set cfg(width) [expr $i - 1]
  $w delete $line "end -1 chars"
  $w yview moveto $yview_top
  $w configure -wrap word -state disabled
}

proc TextScrollWidget { parent title height width } {
  global cfg
  
  label     $parent.label -text $title -font $cfg(titleFont)
  pack      $parent.label -pady 0 -ipady 0 -padx 0 -ipadx 0 -fill x -side top
   
  scrollbar $parent.sbar  -command "$parent.text yview"
  pack      $parent.sbar  -pady 0 -ipady 0 -padx 0 -ipadx 0 -side right -fill y

  text      $parent.text  -relief sunken -wrap word -state disabled \
                          -font $cfg(textFont) -height $height -width $cfg(width)
  # bind all resize events to the script 'FigureOutWidth'
  bind      $parent.text  <Configure> { FigureOutWidth %W }                         
  pack      $parent.text  -pady 0 -ipady 0 -padx 0 -ipadx 0 -side left \
                          -expand 1 -fill both
            $parent.text   configure -yscrollcommand "$parent.sbar set"

}

proc CmdEntryWidget { parent labeltext } {
  global cfg entryprompt smackmode

  label  $parent.label -font $cfg(infoFont) -text $labeltext \
                       -textvariable entryprompt
  entry  $parent.entry -font $cfg(textFont) -width 20 \
                       -relief sunken -bd 2 -textvariable text
  bind   $parent.entry <Delete>     { tkEntryBackspace %W }
  bind   $parent.entry <Return>     { UserInput $text }
  bind   $parent.entry <Up>         { UpHistory }
  bind   $parent.entry <Control-p>  { UpHistory }
  bind   $parent.entry <Down>       { DownHistory }
  bind   $parent.entry <Control-n>  { DownHistory }
  bind   $parent.entry <Tab>        { CommandCompletion; break }
  bind   $parent.entry <Double-Tab> { ListCompletions; break }
  bind   $parent.entry <Control-c>  { %W delete 0 end }
  bind   $parent.entry <Control-l>  { clear "" }
  bind   $parent.entry <Control-t>  { ToggleSmackMode; break }
  bind   $parent.entry <Control-m>  { ToggleMasqueradeMode; break }
  bind   $parent.entry <Control-KeyPress-1>  { set smackmode 0; SetEntryMode; break; }
  bind   $parent.entry <Control-KeyPress-2>  { set smackmode 1; SetEntryMode; break; }
  bind   $parent.entry <Control-KeyPress-3>  { set smackmode 2; SetEntryMode; break; }
  bind   $parent.entry <Control-KeyPress-4>  { set smackmode 3; SetEntryMode; break; }
  bind   $parent.entry <Control-KeyPress-5>  { set smackmode 4; SetEntryMode; break; }

  pack   $parent.label -side left
  pack   $parent.entry -side right -fill x -expand 1
}

proc ToggleSmackMode { } {
  global smackmode

  if { $smackmode == 1 } {
    set smackmode 0
  } else {
    set smackmode 1
  } 
  SetEntryMode
}

proc ToggleMasqueradeMode { } {
  global smackmode

  if { $smackmode == 4 } {
    set smackmode 0
  } else {
    set smackmode 4
  } 
  SetEntryMode
}

proc SetEntryMode { } {
  global smackmode entry entryprompt

  if { $smackmode == 1 } {
    set entryprompt "Talk Smack > "
    bind $entry.entry <Tab> { break }
  } elseif { $smackmode == 2 } {
    set entryprompt "Team Smack > "
    bind $entry.entry <Tab> { break }
  } elseif { $smackmode == 3 } {
    set entryprompt "Self Smack > "
    bind $entry.entry <Tab> { break }
  } elseif { $smackmode == 4 } {
    set entryprompt "Masquerade > "
    bind $entry.entry <Tab> { break }
  } else {
    set entryprompt "Command > "
    bind $entry.entry <Tab> { CommandCompletion; break }
  }
}

proc MakeMainWindow { } {
  global cfg server smack command entry statusbar defbg

  frame .statusbar
  frame .pinfo
  frame .server
  frame .smack
  frame .command
  frame .entry

  StatusBarWidget   .statusbar
  PlayerInfoWidgets .pinfo
  TextScrollWidget  .server    "Server Output"  $cfg(serverWindowHeight)  $cfg(width)
  TextScrollWidget  .smack     "Smack Output"   $cfg(smackWindowHeight)   $cfg(width)
  TextScrollWidget  .command   "Command Output" $cfg(commandWindowHeight) $cfg(width)
  CmdEntryWidget    .entry     "Command > "

  set statusbar .statusbar
  set server    .$cfg(serverPrints)
  set smack     .$cfg(smackPrints)
  set command   .$cfg(commandPrints)
  set entry     .entry

  pack .statusbar -side top    -fill x -expand 0
  pack .pinfo     -side top    -fill x -expand 0
  pack .entry     -side bottom -fill x -expand 0
  if { $cfg(commandWindowHeight) > 0 } {
    pack .command -side bottom -fill both -expand 1
  }
  if { $cfg(serverWindowHeight) > 0 } {
    pack .server  -side top    -fill both -expand 1
  }
  if { $cfg(smackWindowHeight) > 0 } {
    pack .smack   -side top    -fill both -expand 1
  }
  focus  $entry.entry

  set defbg [. cget -bg]
}

proc ExecInitCmds { } {
  global initcmds

  set i 0
  while { [info exists initcmds($i)] } {
    ExecStuffAliasCommand $initcmds($i)
    incr i
  }
}

proc ConnectDefaultQSmack { } {
  global cfg

  if { $cfg(defaultQSmackServer) != "" } {
    UserInput "connect $cfg(defaultQSmackServer)"
  }
}

ReadConfigFile
InitVars
ResetVars
MakeMainWindow
LoadQSmackExtras
ExecInitCmds
ConnectDefaultQSmack
